from dataclasses import dataclass
from typing import *

from .typings import *
from ..typings import ModelInstance

__all__ = [
    "ApplyHandler"
]

from ..utils import iteration


@dataclass
class ApplyHandler:
    instance: Any
    handlers: HandlerDefinitionType
    
    def handle(self) -> AppliedHandlersType:
        fields: ApplyHandlerDefinitionType = {}
        
        for field, handler_list in self.handlers.items():
            for handler in handler_list:
                # Get current value, either from the instance or from previously handled handlers
                current_value = fields.get(field, getattr(self.instance, field))
                # Get new value
                new_value = handler.handle(current_value)
                # Safe new value
                fields[field] = new_value
        
        return fields



from collections import defaultdict

from django_common_utils.libraries.handlers import HandlerDefinitionType
from django_common_utils.libraries.utils import ensure_dict

__all__ = [
    "combine_fields"
]


def combine_fields(first: dict, second: dict) -> dict:
    data = defaultdict(list, first)
    
    for key, value in ensure_dict(second, str, str):
        if value not in data[key]:
            data[key].append(value)
    
    return data



from django.conf import settings
from django.contrib.auth.models import User
from django.template import Context, Template
from django.test import TestCase
from typing import *

from django_common_utils.libraries.utils import  generate_image, model_verbose
from django_common_utils.libraries.utils.common import combine_fields


class LibrariesTest(TestCase):
    @staticmethod
    def render_template(html: str, context: Optional[Dict[str, Any]] = None):
        context = context or {}
        context = Context(context)
        return Template(html).render(context)
    
    def test_utils(self):
        # Common
        self.assertEqual(
            combine_fields({
                "field_1": ["a", "b"],
                "field_2": ["c", "d"]
            }, {
                "field_1": ["a", "b", "c"],
                "field_2": ["e"]
            }),
            {
                "field_1": ["a", "b", "c"],
                "field_2": ["c", "d", "e"]
            }
        )

        # Image
        self.assertEqual(
            generate_image(url="/my_image.jpg"),
            '<img src="/my_image.jpg" alt="" loading="lazy" style="background-color:rgba(0, 0, 0, 0)"  />'
        )
        self.assertHTMLEqual(
            generate_image(url="/my_image.jpg", attributes={
                "data-test": "test_value"
            }),
            '<img src="/my_image.jpg" alt="" loading="lazy" style="background-color:rgba(0, 0, 0, '
            '0)" data-test="test_value" />'
        )
        
        # Text
        assert model_verbose(User) == model_verbose("auth.User") == model_verbose(User.objects.all()) == \
               model_verbose(settings.AUTH_USER_MODEL)
        
    def test_templatetags(self):
        first_html = """
            {% load exceptions math %}
            {% try %}
                {% divide 23 5 %}
            {% except ZeroDivisionError %}
                Catch block (will not be visible)
            {% endtry %}
            {% try %}
                {% divide 23 0 %}
            {% except ZeroDivisionError %}
                Catch block (one exceptions)
            {% endtry %}
            {% try %}
                {% divide 23 0 %}
            {% except %}
                Catch block (all exceptions)
            {% endtry %}
        """
        second_html = """
            {% load exceptions %}
            {% try %}
                {% divide 23 0 %}
            {% except AnExceptionThatHopefullyDoesNotExist %}
                Catch block (will raise error in Python)
            {% endtry %}
        """
        
        expected_first_html = """
            
            
                4
            
            
                Catch block (one exceptions)
            
            
                Catch block (all exceptions)
            
        """
        
        self.assertEqual(self.render_template(first_html), expected_first_html)
        self.assertRaises(ZeroDivisionError, lambda: self.render_template(second_html))
 



import re
from typing import *

from django.template import Context, Node, NodeList, TemplateSyntaxError
from django.template.base import FILTER_SEPARATOR, Parser, Token
from django.template.defaulttags import register
from django.utils.safestring import mark_safe


class ForNode(Node):
    child_nodelists = ('nodelist_loop', 'nodelist_empty')

    def __init__(self, loopvars, sequence, is_reversed, nodelist_loop, nodelist_empty=None):
        self.loopvars, self.sequence = loopvars, sequence
        self.is_reversed = is_reversed
        self.nodelist_loop = nodelist_loop
        if nodelist_empty is None:
            self.nodelist_empty = NodeList()
        else:
            self.nodelist_empty = nodelist_empty

    def render(self, context):
        if 'forloop' in context:
            parentloop = context['forloop']
        else:
            parentloop = {}
        with context.push():
            values = self.sequence.resolve(context, ignore_failures=True)
            if values is None:
                values = []
            if not hasattr(values, '__len__'):
                values = list(values)
            len_values = len(values)
            # Empty render
            if len_values < 1:
                return self.nodelist_empty.render(context)
            nodelist = []
            if self.is_reversed:
                values = reversed(values)
            num_loopvars = len(self.loopvars)
            unpack = num_loopvars > 1
            # Create a forloop value in the context.  We'll update counters on each
            # iteration just below.
            loop_dict = context['forloop'] = {'parentloop': parentloop}
            for i, item in enumerate(values):
                # Shortcuts for current loop iteration number.
                loop_dict['counter0'] = i
                loop_dict['counter'] = i + 1
                # Reverse counter iteration numbers.
                loop_dict['revcounter'] = len_values - i
                loop_dict['revcounter0'] = len_values - i - 1
                # Boolean values designating first and last times through loop.
                loop_dict['first'] = (i == 0)
                loop_dict['last'] = (i == len_values - 1)

                pop_context = False
                if unpack:
                    # If there are multiple loop variables, unpack the item into
                    # them.
                    try:
                        len_item = len(item)
                    except TypeError:  # not an iterable
                        len_item = 1
                    # Check loop variable count before unpacking
                    if num_loopvars != len_item:
                        raise ValueError(
                            "Need {} values to unpack in for loop; got {}. "
                            .format(num_loopvars, len_item),
                        )
                    unpacked_vars = dict(zip(self.loopvars, item))
                    pop_context = True
                    context.update(unpacked_vars)
                else:
                    context[self.loopvars[0]] = item

                for node in self.nodelist_loop:
                    nodelist.append(node.render_annotated(context))

                if pop_context:
                    # Pop the loop variables pushed on to the context to avoid
                    # the context ending up in an inconsistent state when other
                    # tags (e.g., include and with) push data to context.
                    context.pop()
        return mark_safe(''.join(nodelist))



@register.tag('fake_for')
def do_for(parser, token):
    # Ensures format
    bits = token.split_contents()
    if len(bits) < 4:
        raise TemplateSyntaxError("'for' statements should have at least four"
                                  " words: %s" % token.contents)

    is_reversed = bits[-1] == 'reversed'
    in_index = -3 if is_reversed else -2
    if bits[in_index] != 'in':
        raise TemplateSyntaxError("'for' statements should use the format"
                                  " 'for x in y': %s" % token.contents)

    invalid_chars = frozenset((' ', '"', "'", FILTER_SEPARATOR))
    loopvars = re.split(r' *, *', ' '.join(bits[1:in_index]))
    for var in loopvars:
        if not var or not invalid_chars.isdisjoint(var):
            raise TemplateSyntaxError("'for' tag received an invalid argument:"
                                      " %s" % token.contents)

    sequence = parser.compile_filter(bits[in_index + 1])
    nodelist_loop = parser.parse(('empty', 'endfor',))
    token = parser.next_token()
    if token.contents == 'empty':
        # Empty tag
        nodelist_empty = parser.parse(('endfor',))
        parser.delete_first_token()
    else:
        nodelist_empty = None
    return ForNode(loopvars, sequence, is_reversed, nodelist_loop, nodelist_empty)


class TryNode(Node):
    nodelist_try: NodeList
    nodelist_catch: Optional[NodeList]
    child_nodelists = ("nodelist_try", "nodelist_except")
    
    def __init__(
            self,
            nodelist_try: NodeList,
            nodelist_except: Optional[NodeList],
            catch_exceptions: Optional[List[str]] = None
    ):
        self.nodelist_try = nodelist_try
        self.nodelist_except = nodelist_except or NodeList()
        self.catch_exceptions = catch_exceptions
    
    def render(self, context: Context):
        nodelist = []
        
        with context.push():
            try:
                try_data = self.nodelist_try.render(context)
            except Exception as e:
                if \
                        type(self.catch_exceptions) is list and (
                                len(self.catch_exceptions) == 0
                                or e.__class__.__name__ in self.catch_exceptions
                        ):
                    nodelist.append(
                        self.nodelist_except.render(context)
                    )
                else:
                    raise e
            else:
                nodelist.append(
                    try_data
                )
                
        return mark_safe("".join(nodelist))
                    
                


@register.tag("try")
def try_func(parser: Parser, token: Token):
    nodelist_try: NodeList
    nodelist_except: Optional[NodeList] = None
    catch_exceptions: Optional[List] = None
    
    # Try
    nodelist_try: Token = parser.parse(("except", "endtry"))
    
    next_token = parser.next_token()
    next_token_name = next_token.contents.lstrip().split(" ")[0].lower()
    
    # Catch
    if next_token_name == "except":
        # Catch
        contents = next_token.contents.split(" ")[1:]
        
        if len(contents) > 0:
            catch_exceptions = contents
        else:
            catch_exceptions = []
        
        nodelist_except = parser.parse(("endtry", ))
    else:
        parser.parse(("endtry",))
    
    parser.next_token()
    
    return TryNode(
        nodelist_try=nodelist_try,
        nodelist_except=nodelist_except,
        catch_exceptions=catch_exceptions,
    )




import inspect
import re
from typing import *

from django.db import models
from django.db.models import QuerySet
from django.utils.html import strip_tags
from django.utils.translation import gettext_lazy as _
from django.apps import apps

from ..typings import ModelInstance


def get_model(value) -> ModelInstance:
    if type(value) is str:
        return apps.get_model(*value.split(".", 1), require_ready=False)
    
    if isinstance(value, QuerySet):
        return value.model
    
    if inspect.isclass(value) and issubclass(value, models.Model):
        return value
    
    if issubclass(value.__class__, models.Model):
        return value

    raise ValueError("Model not found")


def create_short(text: str, max_length: int, prefix: str = "...") -> str:
    return f"{text[:max_length]}{prefix}" if len(text) > max_length else text


def model_verbose(model) -> str:
    # noinspection PyProtectedMember
    return get_model(model)._meta.verbose_name


def model_verbose_plural(model) -> str:
    # noinspection PyProtectedMember
    return get_model(model)._meta.verbose_name_plural


def camelcase_to_underscore(value: str) -> str:
    x: str
    return "_".join(
        [x.lower() for x in re.findall("[A-Z][^A-Z]*", value)]
    )


def listify(x: List[str], /) -> str:
    length = len(x)
    
    if length == 0:
        return ""
    elif length == 1:
        return x[0]
    return _("{} and {}").format(
        ", ".join([str(value) for value in x[:-1]]),
        x[-1]
    )


def field_verbose(model: ModelInstance, field: str) -> str:
    return model._meta.get_field(field).verbose_name


def field_verbose_plural(model: ModelInstance, field: str) -> str:
    return model._meta.get_field(field).verbose_name_plural


def textify(html: str) -> str:
    """Textifies an html input"""
    # Remove html tags and continuous whitespaces
    text_only = re.sub("[ \t]+", " ", strip_tags(html))
    # Strip single spaces in the beginning of each line
    return text_only.replace('\n ', '\n').strip()



from django_common_utils.libraries.fieldsets import AdminFieldsetMixin
from django_common_utils.libraries.utils import common


class CreationDateAdminFieldsetMixin(AdminFieldsetMixin):
    def get_mixin_fields(self, **_) -> dict:
        return {
            "created": "created_at"
        }
    
    def get_readonly_fields(self, **_):
        return "created_at"
    

class EditCreationDateAdminFieldsetMixin(AdminFieldsetMixin):
    def get_mixin_fields(self, **kwargs):
        data = super().get_mixin_fields(**kwargs)
        own_data = {
            "created": "edited_at"
        }
        
        return common.combine_fields(data, own_data)
    
    def get_readonly_fields(self, **_):
        return "edited_at"



from datetime import datetime, timedelta
from typing import *

from ..admin import AdminFieldsetMixin


class SlugAdminFieldsetMixin(AdminFieldsetMixin):
    def get_readonly_fields(self, obj=None, **_):
        if obj:
            # If the object has a pub_date, make slug only readonly if the obj's publish date or edited date is older
            # than the given threshold.
            try_dates: List[str] = ["pub_date", "edited_at", "edit_date"]
            time = datetime.now() - timedelta(hours=obj._COMMON_SLUG_CHANGE_THRESHOLD)
            
            for date in try_dates:
                if (value := getattr(obj, date, None)) is not None:
                    if value > time:
                        return
            return "slug"
    
    def get_mixin_fields(self, **_) -> dict:
        return {
            "default": ["slug", ]
        }
